# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Library for working with projects."""

_beta_test_mode = False


def _enter_beta_test_mode():
    """Called by anaconda-project executable to do special things for beta."""
    global _beta_test_mode
    _beta_test_mode = True



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Handles uncaught exceptions."""
from __future__ import absolute_import, print_function

import sys

import anaconda_project.internal.cli.console_utils as console_utils
from anaconda_project.internal.slugify import slugify


def handle_bugs(main_func, program_name, details_dict):
    """Invoke main entry point, handling uncaught exceptions.

    Args:
        main_func (function): a main()-style function returning an exit status
        program_name (str): name of the app
        details_dict (dict): dictionary of stuff to include in the bug report file

    Returns:
        an exit status code from main_func or 1 on uncaught exception
    """
    slugified_program_name = slugify(program_name)

    try:
        return main_func()
    except KeyboardInterrupt:
        # KeyboardInterrupt doesn't derive from Exception, but the default handler
        # prints a stack trace which is sort of ugly, so we replace the default
        # with this.
        print("%s was interrupted." % program_name, file=sys.stderr)
        return 1
    except Exception:
        (exception_type, exception_value, exception_trace) = sys.exc_info()

        try:
            import datetime
            import pprint
            import tempfile
            import traceback

            print("An unexpected error occurred, most likely a bug in %s." % program_name, file=sys.stderr)
            print("    (The error was: %s: %s)" % (exception_type.__name__, str(exception_value)), file=sys.stderr)

            # so batch jobs have the details in their logs
            output_to_console = not console_utils.stdin_is_interactive()

            when = datetime.date.today().isoformat()
            prefix = "bug_details_%s_%s_" % (slugified_program_name, when)
            with tempfile.NamedTemporaryFile(prefix=prefix, suffix=".txt", delete=False) as bugfile:
                report_name = bugfile.name

                def output(s):
                    bugfile.write(s.encode('utf-8'))
                    bugfile.write("\n".encode('utf-8'))
                    if output_to_console:
                        print(s, file=sys.stderr)

                output("Bug details for %s error on %s" % (program_name, when))
                output("")
                output("sys.argv: %r" % sys.argv)
                output("")
                output(pprint.pformat(details_dict))
                output("")
                output("\n".join(traceback.format_exception(exception_type, exception_value, exception_trace)))

            if output_to_console:
                print("Above details were also saved to %s" % report_name, file=sys.stderr)
            else:
                print("Details about the error were saved to %s" % report_name, file=sys.stderr)
        except Exception:
            # re-raise the original exception, which is probably more useful
            # than reporting whatever was broken about our bug handling code
            # above.
            raise exception_value

        # exit code
        return 1



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""The ``archive`` command makes an archive of the project."""
from __future__ import absolute_import, print_function

from anaconda_project.internal.cli.project_load import load_project
from anaconda_project.internal.cli import console_utils
import anaconda_project.project_ops as project_ops


def archive_command(project_dir, archive_filename):
    """Make an archive of the project.

    Returns:
        exit code
    """
    project = load_project(project_dir)
    status = project_ops.archive(project, archive_filename)
    if status:
        print(status.status_description)
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def main(args):
    """Start the archive command and return exit status code."""
    return archive_command(args.directory, args.filename)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2017, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Analyze notebook files."""
from __future__ import absolute_import

import codecs
import json
import re

from anaconda_project.internal.py2_compat import is_string

_comment_re = re.compile("#.*$", re.MULTILINE)
_fusion_register_re = re.compile(r"^\s*@fusion\.register", re.MULTILINE)


# see if some source has @fusion.register. This is
# obviously sort of heuristic, but without executing
# the python we can only do so much.
def _has_fusion_register(source):
    # dump comments so commenting out fusion.register
    # would work as expected
    source = re.sub(_comment_re, "", source)
    return re.match(_fusion_register_re, source) is not None


def extras(filename, errors):
    try:
        with codecs.open(filename, encoding='utf-8') as f:
            json_string = f.read()
            parsed = json.loads(json_string)
    except Exception as e:
        errors.append("Failed to read or parse %s: %s" % (filename, str(e)))
        return None

    extras = dict()
    found_fusion = False

    if isinstance(parsed, dict) and \
       'cells' in parsed and \
       isinstance(parsed['cells'], list):
        for cell in parsed['cells']:
            if 'source' in cell:
                if isinstance(cell['source'], list):
                    source = "".join([s for s in cell['source'] if is_string(s)])
                    if _has_fusion_register(source):
                        found_fusion = True

    if found_fusion:
        extras['registers_fusion_function'] = True

    return extras



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""File download requirements."""

from __future__ import absolute_import, print_function

import os

from anaconda_project.requirements_registry.requirement import EnvVarRequirement
from anaconda_project.requirements_registry.network_util import urlparse

from anaconda_project.internal.py2_compat import is_string

_hash_algorithms = ('md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512')


class DownloadRequirement(EnvVarRequirement):
    """A requirement for ``env_var`` to point to a downloaded file."""

    @classmethod
    def _parse(cls, varname, item, problems):
        """Parse an item from the downloads: section."""
        url = None
        filename = None
        hash_algorithm = None
        hash_value = None
        unzip = None
        description = None
        if is_string(item):
            url = item
        elif isinstance(item, dict):
            url = item.get('url', None)
            if url is None:
                problems.append("Download item {} doesn't contain a 'url' field.".format(varname))
                return None

            description = item.get('description', None)
            if description is not None and not is_string(description):
                problems.append("'description' field for download item {} is not a string".format(varname))
                return None

            for method in _hash_algorithms:
                if method not in item:
                    continue

                if hash_algorithm is not None:
                    problems.append("Multiple checksums for download {}: {} and {}.".format(varname, hash_algorithm,
                                                                                            method))
                    return None
                else:
                    hash_value = item[method]
                    if is_string(hash_value):
                        hash_algorithm = method
                    else:
                        problems.append("Checksum value for {} should be a string not {}.".format(varname, hash_value))
                        return None

            filename = item.get('filename', None)
            unzip = item.get('unzip', None)
            if unzip is not None and not isinstance(unzip, bool):
                problems.append("Value of 'unzip' for download item {} should be a boolean, not {}.".format(varname,
                                                                                                            unzip))
                return None

        if url is None or not is_string(url):
            problems.append(("Download name {} should be followed by a URL string or a dictionary " +
                             "describing the download.").format(varname))
            return None

        if url == '':
            problems.append("Download item {} has an empty 'url' field.".format(varname))
            return None

        # urlsplit doesn't seem to ever throw an exception, but it can
        # return pretty nonsensical stuff on invalid urls, in particular
        # an empty path is very possible
        url_path = os.path.basename(urlparse.urlsplit(url).path)
        url_path_is_zip = url_path.lower().endswith(".zip")

        if filename is None:
            if url_path != '':
                filename = url_path
                if url_path_is_zip:
                    if unzip is None:
                        # url is a zip and neither filename nor unzip specified, assume unzip
                        unzip = True
                    if unzip:
                        # unzip specified True, or we guessed True, and url ends in zip;
                        # take the .zip off the filename we invented based on the url.
                        filename = filename[:-4]
        elif url_path_is_zip and unzip is None and not filename.lower().endswith(".zip"):
            # URL is a zip, filename is not a zip, unzip was not specified, so assume
            # we want to unzip
            unzip = True

        if filename is None:
            filename = varname

        if unzip is None:
            unzip = False

        return dict(env_var=varname,
                    url=url,
                    filename=filename,
                    hash_algorithm=hash_algorithm,
                    hash_value=hash_value,
                    unzip=unzip,
                    description=description)

    def __init__(self,
                 registry,
                 env_var,
                 url,
                 filename,
                 hash_algorithm=None,
                 hash_value=None,
                 unzip=False,
                 description=None):
        """Extend init to accept url and hash parameters."""
        options = None
        if description is not None:
            options = dict(description=description)
        super(DownloadRequirement, self).__init__(registry=registry, env_var=env_var, options=options)
        assert url is not None
        assert filename is not None
        assert len(url) > 0
        assert len(filename) > 0
        self.url = url
        self.filename = filename
        assert hash_algorithm is None or hash_algorithm in _hash_algorithms
        self.hash_algorithm = hash_algorithm
        self.hash_value = hash_value
        self.unzip = unzip

    @property
    def description(self):
        """Override superclass to supply our description."""
        return self._description("A downloaded file which is referenced by {}.".format(self.env_var))

    @property
    def ignore_patterns(self):
        """Override superclass with our ignore patterns."""
        return set(['/' + self.filename, '/' + self.filename + ".part"])

    def _why_not_provided(self, environ):
        if self.env_var not in environ:
            return self._unset_message()
        filename = environ[self.env_var]
        if not os.path.exists(filename):
            return 'File not found: {}'.format(filename)

    def check_status(self, environ, local_state_file, default_env_spec_name, overrides, latest_provide_result=None):
        """Override superclass to get our status."""
        why_not_provided = self._why_not_provided(environ)

        has_been_provided = why_not_provided is None
        if has_been_provided:
            status_description = ("File downloaded to {}".format(self._get_value_of_env_var(environ)))
        else:
            status_description = why_not_provided
        return self._create_status(environ,
                                   local_state_file,
                                   default_env_spec_name,
                                   overrides=overrides,
                                   has_been_provided=has_been_provided,
                                   status_description=status_description,
                                   provider_class_name='DownloadProvider',
                                   latest_provide_result=latest_provide_result)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import collections


class CycleError(Exception):
    def __init__(self, involving):
        message = "Cycle in graph involving {involving}".format(involving=involving)
        super(CycleError, self).__init__(message)
        self.involving = involving


def toposort(nodes, get_next_nodes):
    """Sort list of graph nodes.

    Returns a new list, does not modify input list.

    Args:
        nodes (iterable): iterable of some kind of node
        get_next_nodes (function): takes a node and returns iterable of next nodes

    Returns:
        new sorted list of nodes
    """
    traversing = set()
    traversed = set()
    result = collections.deque()

    def traverse(node):
        if node in traversing:
            raise CycleError(node)
        if node in traversed:
            return  # not a cycle but we already saw this
        traversing.add(node)
        for next in get_next_nodes(node):
            traverse(next)
        traversed.add(node)
        traversing.remove(node)
        result.appendleft(node)

    for node in nodes:
        traverse(node)

    return list(result)


def toposort_from_dependency_info(nodes, get_node_key, get_dependency_keys, can_ignore_dependency=None):
    """Sort list of nodes that depend on other nodes in dependency-first order.

    All dependencies must be in the list of nodes.

    Returns a new list, does not modify input list.

    Args:
        nodes (iterable): iterable of some kind of node
        get_node_key (function): get identifier for a node
        get_dependency_keys (function): get iterable of node identifiers a node depends on

    Returns:
        new sorted list of nodes
    """
    nodes_by_key = dict()
    node_depended_on_by = dict()

    for node in nodes:
        key = get_node_key(node)
        if key in nodes_by_key:
            raise ValueError("two nodes with the same key %r" % key)
        nodes_by_key[key] = node
        node_depended_on_by[key] = set()

    for node in nodes:
        dep_keys = get_dependency_keys(node)
        for dep_key in dep_keys:
            if dep_key not in nodes_by_key:
                if can_ignore_dependency is None or not can_ignore_dependency(dep_key):
                    raise ValueError("Dependency %r was not in the list of nodes %r" % (dep_key, nodes))
            else:
                node_depended_on_by[dep_key].add(node)

    return toposort(nodes, lambda n: node_depended_on_by[get_node_key(n)])



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""The ``main`` function chooses and runs a subcommand."""
from __future__ import absolute_import, print_function

# the point of this file is to make the internal main() into a public
# entry point.
import anaconda_project.internal.cli.main as cli_main


def main():
    """anaconda-project command line tool Conda-style entry point.

    Conda expects us to take no args and return an exit code.
    """
    return cli_main.main()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import subprocess

from anaconda_project import verbose


def _log_args(args):
    log = verbose._verbose_logger()
    log.info("$ %s", " ".join(args))


def call(args, **kwargs):
    _log_args(args)
    return subprocess.call(args=args, **kwargs)


def Popen(args, **kwargs):
    _log_args(args)
    return subprocess.Popen(args=args, **kwargs)


def check_output(args, **kwargs):
    _log_args(args)
    return subprocess.check_output(args=args, **kwargs)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from anaconda_project.internal.py2_compat import is_unicode

import re
import unicodedata

_remove_chars = re.compile('[^A-Za-z0-9-_]', re.UNICODE)


def slugify(s):
    """A simple slugifier.

    This keeps ascii alphanumerics, -, and _, but replaces
    everything else with hyphen.
    """
    if not is_unicode(s):
        # normalize() requires a unicode string
        s = s.decode(encoding='utf-8', errors='replace')
    s = unicodedata.normalize('NFC', s)

    # The complicating mess here is that "narrow" builds of Python
    # are really UTF-16, not arrays of unicode characters, so we have
    # to deal with surrogate pairs. re.sub, len(), etc. will all treat
    # surrogate pairs as multiple characters. We have to deal with that
    # by hand to avoid a different slug on different platforms.
    def replace(c):
        # ignore the first half of any surrogate pair, the
        # second half will become a hyphen.
        if 0xD800 <= ord(c[0]) <= 0xDBFF:
            return ""  # pragma: no cover (only on "narrow" python builds)
        elif _remove_chars.match(c):
            return "-"
        else:
            return c

    return "".join(map(replace, s))



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import errno
import os


def makedirs_ok_if_exists(path):
    try:
        os.makedirs(path)
    except IOError as e:  # pragma: no cover (py3 only)
        if e.errno != errno.EEXIST:
            raise e
    except OSError as e:  # pragma: no cover (py2 only)
        if e.errno != errno.EEXIST:
            raise e
    return path



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Bundle up a project for shipment."""
from __future__ import absolute_import, print_function

import codecs
import errno
import fnmatch
import os
import platform
import shutil
import subprocess
import tarfile
import tempfile
import uuid
import zipfile

from anaconda_project.frontend import _new_error_recorder
from anaconda_project.internal import logged_subprocess
from anaconda_project.internal.simple_status import SimpleStatus
from anaconda_project.internal.directory_contains import subdirectory_relative_to_directory
from anaconda_project.internal.rename import rename_over_existing
from anaconda_project.internal.makedirs import makedirs_ok_if_exists


class _FileInfo(object):
    def __init__(self, project_directory, filename, is_directory):
        self.full_path = os.path.abspath(filename)
        self.relative_path = os.path.relpath(self.full_path, start=project_directory)
        if platform.system() == 'Windows':
            self.unixified_relative_path = self.relative_path.replace("\\", "/")
        else:
            self.unixified_relative_path = self.relative_path
        self.basename = os.path.basename(self.full_path)
        self.is_directory = is_directory


def _list_project(project_directory, ignore_filter, frontend):
    try:
        file_infos = []
        for root, dirs, files in os.walk(project_directory):
            filtered_dirs = []
            for d in dirs:
                info = _FileInfo(project_directory=project_directory, filename=os.path.join(root, d), is_directory=True)
                if ignore_filter(info):
                    continue
                else:
                    filtered_dirs.append(d)
                    file_infos.append(info)

            # don't even recurse into filtered-out directories, mostly because recursing into
            # "envs" is very slow
            dirs[:] = filtered_dirs

            for f in files:
                info = _FileInfo(project_directory=project_directory,
                                 filename=os.path.join(root, f),
                                 is_directory=False)
                if not ignore_filter(info):
                    file_infos.append(info)

        return file_infos
    except OSError as e:
        frontend.error("Could not list files in %s: %s." % (project_directory, str(e)))
        return None


class _FilePattern(object):
    def __init__(self, pattern):
        assert pattern != ''
        # the glob string
        self.pattern = pattern

    def matches(self, info):
        # Unlike .gitignore, this is a path-unaware match; fnmatch doesn't pay
        # any attention to "/" as a special character. However, on Windows, we
        # have fixed up unixified_relative_path to have / instead of \, so that
        # it will match patterns specified with /.
        def match(path, pattern):
            assert path.startswith("/")
            while path != '/':
                if fnmatch.fnmatch(path, pattern):
                    return True
                # this assumes that on Windows, dirname on a unixified path
                # will do the right thing...
                path = os.path.dirname(path)
            return False

        if self.pattern.startswith("/"):
            # we have to match the full path or one of its parents exactly
            pattern = self.pattern
        else:
            # we only have to match the end of the path (implicit "*/")
            pattern = "*/" + self.pattern

        # So that */ matches even plain "foo" we need to start with /
        match_against = "/" + info.unixified_relative_path

        # ending with / means only match directories
        if pattern.endswith("/"):
            if info.is_directory:
                return match(match_against, pattern[:-1])
            else:
                return False
        else:
            return match(match_against, pattern)


def _parse_ignore_file(filename, frontend):
    patterns = []
    try:
        with codecs.open(filename, 'r', 'utf-8') as f:
            for line in f:
                line = line.strip()

                # comments can't be at the end of the line,
                # you can only comment out an entire line.
                if line.startswith("#"):
                    continue

                # you can backslash-escape a hash at the start
                if line.startswith("\#"):
                    line = line[1:]

                if line != '':
                    patterns.append(_FilePattern(pattern=line))
        return patterns
    except (OSError, IOError) as e:
        # it's ok for .projectignore to be absent, but not OK
        # to fail to read it if present.
        if e.errno == errno.ENOENT:
            # return default patterns anyway
            return patterns
        else:
            frontend.error("Failed to read %s: %s" % (filename, str(e)))
            return None


def _load_ignore_file(project_directory, frontend):
    ignore_file = os.path.join(project_directory, ".projectignore")
    return _parse_ignore_file(ignore_file, frontend)


def _git_ignored_files(project_directory, frontend):
    if not os.path.exists(os.path.join(project_directory, ".git")):
        return []

    # It is pretty involved to parse .gitignore correctly. Lots of
    # little syntax rules that don't quite match python's fnmatch,
    # there can be multiple .gitignore, and there are also things
    # in the git config file that affect what's ignored.  So we
    # let git do this itself. If the project has a `.git` we assume
    # the user is using git.

    # --other means show untracked (not added) files
    # --ignored means show ignored files
    # --exclude-standard means use the usual .gitignore and other configuration
    # --directory means output "node_modules/" if it's ignored, not 100000 JS files
    try:
        output = logged_subprocess.check_output(
            ['git', 'ls-files', '--others', '--ignored', '--exclude-standard', '--directory'],
            cwd=project_directory)
        # for whatever reason, git doesn't include the ".git" in the ignore list
        return [".git"] + output.decode('utf-8').splitlines()
    except subprocess.CalledProcessError as e:
        message = e.output.decode('utf-8').replace("\n", " ")
        frontend.error("'git ls-files' failed to list ignored files: %s." % (message))
        return None
    except OSError as e:
        frontend.error("Failed to run 'git ls-files'; %s" % str(e))
        return None


def _git_filter(project_directory, frontend):
    git_ignored = _git_ignored_files(project_directory, frontend)
    if git_ignored is None:
        return None

    git_ignored = set(git_ignored)

    def is_git_ignored(info):
        path = info.relative_path
        while path != '':
            assert path != '/'  # would infinite loop
            # git ls-files seems to append "/" to dirs when using --directory
            if path in git_ignored or (path + '/') in git_ignored:
                return True
            path = os.path.dirname(path)
        return False

    return is_git_ignored


def _ignore_file_filter(project_directory, frontend):
    patterns = _load_ignore_file(project_directory, frontend)
    if patterns is None:
        return None

    def matches_some_pattern(info):
        for pattern in patterns:
            if pattern.matches(info):
                return True
        return False

    return matches_some_pattern


def _enumerate_archive_files(project_directory, frontend, requirements):
    git_filter = _git_filter(project_directory, frontend)
    ignore_file_filter = _ignore_file_filter(project_directory, frontend)
    if git_filter is None or ignore_file_filter is None:
        return None

    plugin_patterns = set()
    for req in requirements:
        plugin_patterns = plugin_patterns.union(req.ignore_patterns)
    plugin_patterns = [_FilePattern(s) for s in plugin_patterns]

    def is_plugin_generated(info):
        for pattern in plugin_patterns:
            if pattern.matches(info):
                return True
        return False

    def all_filters(info):
        return git_filter(info) or ignore_file_filter(info) or is_plugin_generated(info)

    infos = _list_project(project_directory, all_filters, frontend)
    if infos is None:
        return None

    return infos


def _leaf_infos(infos):
    all_by_name = dict()
    for info in infos:
        all_by_name[info.relative_path] = info
    for info in infos:
        parent = os.path.dirname(info.relative_path)
        while parent != '':
            assert parent != '/'  # would be infinite loop
            if parent in all_by_name:
                del all_by_name[parent]
            parent = os.path.dirname(parent)

    return sorted(all_by_name.values(), key=lambda x: x.relative_path)


def _write_tar(archive_root_name, infos, filename, compression, frontend):
    if compression is None:
        compression = ""
    else:
        compression = ":" + compression
    with tarfile.open(filename, ('w%s' % compression)) as tf:
        for info in _leaf_infos(infos):
            arcname = os.path.join(archive_root_name, info.relative_path)
            frontend.info("  added %s" % arcname)
            tf.add(info.full_path, arcname=arcname)


def _write_zip(archive_root_name, infos, filename, frontend):
    with zipfile.ZipFile(filename, 'w') as zf:
        for info in _leaf_infos(infos):
            arcname = os.path.join(archive_root_name, info.relative_path)
            frontend.info("  added %s" % arcname)
            zf.write(info.full_path, arcname=arcname)


# function exported for project.py
def _list_relative_paths_for_unignored_project_files(project_directory, frontend, requirements):
    infos = _enumerate_archive_files(project_directory, frontend, requirements=requirements)
    if infos is None:
        return None
    return [info.relative_path for info in infos]


# function exported for project_ops.py
def _archive_project(project, filename):
    """Make an archive of the non-ignored files in the project.

    Args:
        project (``Project``): the project
        filename (str): name for the new zip or tar.gz archive file

    Returns:
        a ``Status``, if failed has ``errors``
    """
    failed = project.problems_status()
    if failed is not None:
        for error in failed.errors:
            project.frontend.error(error)
        return failed

    frontend = _new_error_recorder(project.frontend)

    if not os.path.exists(project.project_file.filename):
        frontend.error("%s does not exist." % project.project_file.basename)
        return SimpleStatus(success=False, description="Can't create an archive.", errors=frontend.pop_errors())

    # this would most likely happen in a GUI editor, if it reloaded
    # the project from memory but hadn't saved yet.
    if project.project_file.has_unsaved_changes:
        frontend.error("%s has been modified but not saved." % project.project_file.basename)
        return SimpleStatus(success=False, description="Can't create an archive.", errors=frontend.pop_errors())

    infos = _enumerate_archive_files(project.directory_path,
                                     frontend,
                                     requirements=project.union_of_requirements_for_all_envs)
    if infos is None:
        return SimpleStatus(success=False,
                            description="Failed to list files in the project.",
                            errors=frontend.pop_errors())

    # don't put the destination zip into itself, since it's fairly natural to
    # create a archive right in the project directory
    relative_dest_file = subdirectory_relative_to_directory(filename, project.directory_path)
    if not os.path.isabs(relative_dest_file):
        infos = [info for info in infos if info.relative_path != relative_dest_file]

    tmp_filename = filename + ".tmp-" + str(uuid.uuid4())
    try:
        if filename.lower().endswith(".zip"):
            _write_zip(project.name, infos, tmp_filename, frontend)
        elif filename.lower().endswith(".tar.gz"):
            _write_tar(project.name, infos, tmp_filename, compression="gz", frontend=frontend)
        elif filename.lower().endswith(".tar.bz2"):
            _write_tar(project.name, infos, tmp_filename, compression="bz2", frontend=frontend)
        elif filename.lower().endswith(".tar"):
            _write_tar(project.name, infos, tmp_filename, compression=None, frontend=frontend)
        else:
            frontend.error("Unsupported archive filename %s." % (filename))
            return SimpleStatus(success=False,
                                description="Project archive filename must be a .zip, .tar.gz, or .tar.bz2.",
                                errors=frontend.pop_errors())
        rename_over_existing(tmp_filename, filename)
    except IOError as e:
        frontend.error(str(e))
        return SimpleStatus(success=False,
                            description=("Failed to write project archive %s." % (filename)),
                            errors=frontend.pop_errors())
    finally:
        try:
            os.remove(tmp_filename)
        except (IOError, OSError):
            pass

    unlocked = []
    for env_spec in project.env_specs.values():
        if env_spec.lock_set.disabled:
            unlocked.append(env_spec.name)

    if len(unlocked) > 0:
        frontend.info("Warning: env specs are not locked, which means they may not "
                      "work consistently for others or when deployed.")
        frontend.info("  Consider using the 'anaconda-project lock' command to lock the project.")
        if len(unlocked) != len(project.env_specs):
            frontend.info("  Unlocked env specs are: " + (", ".join(sorted(unlocked))))

    return SimpleStatus(success=True, description=("Created project archive %s" % filename))


def _list_files_zip(zip_path):
    with zipfile.ZipFile(zip_path, mode='r') as zf:
        return sorted(zf.namelist())


def _list_files_tar(tar_path):
    with tarfile.open(tar_path, mode='r') as tf:
        # we don't want links or block devices or anything weird, they could be a security problem
        return sorted([member.name for member in tf.getmembers() if member.isreg() or member.isdir()])


def _extract_files_zip(zip_path, src_and_dest, frontend):
    # the zipfile API has no way to extract to a filename of
    # our choice, so we have to unpack to a temporary location,
    # then copy those files over.
    tmpdir = tempfile.mkdtemp()
    try:
        with zipfile.ZipFile(zip_path, mode='r') as zf:
            zf.extractall(tmpdir)
            for (src, dest) in src_and_dest:
                frontend.info("Unpacking %s to %s" % (src, dest))
                src_path = os.path.join(tmpdir, src)
                if os.path.isdir(src_path):
                    makedirs_ok_if_exists(dest)
                    shutil.copystat(src_path, dest)
                else:
                    makedirs_ok_if_exists(os.path.dirname(dest))
                    shutil.copy2(src_path, dest)
    finally:
        try:
            shutil.rmtree(tmpdir)
        except (IOError, OSError):
            pass


def _extract_files_tar(tar_path, src_and_dest, frontend):
    with tarfile.open(tar_path, mode='r') as tf:
        for (src, dest) in src_and_dest:
            frontend.info("Unpacking %s to %s" % (src, dest))
            member = tf.getmember(src)
            # we could also use tf._extract_member here, but the
            # solution below with only the public API isn't that
            # bad.
            if member.isreg():
                makedirs_ok_if_exists(os.path.dirname(dest))
                tf.makefile(member, dest)
            else:
                assert member.isdir()  # we filtered out other types
                makedirs_ok_if_exists(dest)

            try:
                tf.chown(member, dest, False)  # pragma: no cover (python 3.5 has another param)
            except TypeError:  # pragma: no cover
                tf.chown(member, dest)  # pragma: no cover (python 2.7, 3.4)
            tf.chmod(member, dest)
            tf.utime(member, dest)


def _split_after_first(path):
    # starting from archive name, be sure we have a valid path for this OS
    path = path.replace("/", os.sep)

    def _helper(head, tail):
        (dirname, filename) = os.path.split(head)
        if dirname == '':
            # head had no separators, so return it as the first
            return (head, tail)
        elif tail is None:
            # we were called with no tail, so create the first tail
            return _helper(dirname, filename)
        else:
            # add filename to the tail
            return _helper(dirname, os.path.join(filename, tail))

    return _helper(path, None)


def _get_source_and_dest_files(archive_path, list_files, project_dir, parent_dir, frontend):

    names = list_files(archive_path)
    if len(names) == 0:
        frontend.error("A valid project archive must contain at least one file.")
        return None
    items = [(name, prefix, remainder)
             for (name, (prefix, remainder)) in zip(names, [_split_after_first(name) for name in names])]
    candidate_prefix = items[0][1]
    if candidate_prefix == "..":
        frontend.error("Archive contains relative path '%s' which is not allowed." % (items[0][0]))
        return None

    if project_dir is None:
        project_dir = candidate_prefix

    if os.path.isabs(project_dir):
        assert parent_dir is None
        canonical_project_dir = os.path.realpath(os.path.abspath(project_dir))
        canonical_parent_dir = os.path.dirname(canonical_project_dir)
    else:
        if parent_dir is None:
            parent_dir = os.getcwd()

        canonical_parent_dir = os.path.realpath(os.path.abspath(parent_dir))
        canonical_project_dir = os.path.realpath(os.path.abspath(os.path.join(canonical_parent_dir, project_dir)))

    # candidate_prefix is untrusted and may try to send us outside of parent_dir.
    # this assertion is because of the check for candidate_prefix == ".." above.
    assert canonical_project_dir.startswith(canonical_parent_dir)

    if os.path.exists(canonical_project_dir):
        # This is an error to ensure we always do a "fresh" unpack
        # without worrying about overwriting stuff.
        frontend.error("Directory '%s' already exists." % canonical_project_dir)
        return None

    src_and_dest = []
    for (name, prefix, remainder) in items:
        if prefix != candidate_prefix:
            frontend.error(("A valid project archive contains only one project directory " +
                            "with all files inside that directory. '%s' is outside '%s'.") % (name, candidate_prefix))
            return None
        if remainder is None:
            # this is an entry that's either the prefix dir itself,
            # or a file at the root not in any dir
            continue
        dest = os.path.realpath(os.path.abspath(os.path.join(canonical_project_dir, remainder)))
        # this check deals with ".." in the name for example
        if not dest.startswith(canonical_project_dir):
            frontend.error("Archive entry '%s' would end up at '%s' which is outside '%s'." %
                           (name, dest, canonical_project_dir))
            return None
        src_and_dest.append((name, dest))

    return (canonical_project_dir, src_and_dest)


class _UnarchiveStatus(SimpleStatus):
    def __init__(self, success, description, project_dir):
        super(_UnarchiveStatus, self).__init__(success=success, description=description)
        self.project_dir = project_dir


# function exported for project_ops.py
def _unarchive_project(archive_filename, project_dir, frontend, parent_dir=None):
    """Unpack an archive of files in the project.

    This takes care of several details, for example it deals with
    hostile archives containing files outside of the dest
    directory, and it handles both tar and zip.

    It does not load or validate the unpacked project.

    project_dir can be None to auto-choose one.

    If parent_dir is non-None, place the project_dir in it. This is most useful
    if project_dir is None.

    Args:
        archive_filename (str): the tar or zip archive file
        project_dir (str): the directory that will contain the project config file
        parent_dir (str): place project directory in here

    Returns:
        a ``Status``, if failed has ``errors``, on success has a ``project_dir`` property
    """
    if project_dir is not None and os.path.isabs(project_dir) and parent_dir is not None:
        raise ValueError("If supplying parent_dir to unarchive, project_dir must be relative or None")

    frontend = _new_error_recorder(frontend)

    list_files = None
    extract_files = None
    if archive_filename.endswith(".zip"):
        list_files = _list_files_zip
        extract_files = _extract_files_zip
    elif any([archive_filename.endswith(suffix) for suffix in [".tar", ".tar.gz", ".tar.bz2"]]):
        list_files = _list_files_tar
        extract_files = _extract_files_tar
    else:
        frontend.error("Unsupported archive filename %s, must be a .zip, .tar.gz, or .tar.bz2" % (archive_filename))
        return SimpleStatus(success=False,
                            description=("Could not unpack archive %s" % archive_filename),
                            errors=frontend.pop_errors())

    try:
        result = _get_source_and_dest_files(archive_filename, list_files, project_dir, parent_dir, frontend)
        if result is None:
            return SimpleStatus(success=False,
                                description=("Could not unpack archive %s" % archive_filename),
                                errors=frontend.pop_errors())
        (canonical_project_dir, src_and_dest) = result

        if len(src_and_dest) == 0:
            frontend.error("Archive does not contain a project directory or is empty.")
            return SimpleStatus(success=False,
                                description=("Could not unpack archive %s" % archive_filename),
                                errors=frontend.pop_errors())

        assert not os.path.exists(canonical_project_dir)
        os.makedirs(canonical_project_dir)

        try:
            extract_files(archive_filename, src_and_dest, frontend)
        except Exception as e:
            try:
                shutil.rmtree(canonical_project_dir)
            except (IOError, OSError):
                pass
            raise e

        return _UnarchiveStatus(success=True,
                                description=("Project archive unpacked to %s." % canonical_project_dir),
                                project_dir=canonical_project_dir)
    except (IOError, OSError, zipfile.error, tarfile.TarError) as e:
        frontend.error(str(e))
        return SimpleStatus(success=False, description="Failed to read project archive.", errors=frontend.pop_errors())



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Commands related to the downloads section."""
from __future__ import absolute_import, print_function

import sys

from anaconda_project.internal.cli.project_load import load_project
from anaconda_project import project_ops
from anaconda_project.internal.cli import console_utils
from anaconda_project.prepare import prepare_without_interaction
from anaconda_project.provide import PROVIDE_MODE_CHECK


def add_download(project_dir, env_spec_name, filename_variable, download_url, filename, hash_algorithm, hash_value):
    """Add an item to the downloads section."""
    project = load_project(project_dir)
    if (hash_algorithm or hash_value) and not bool(hash_algorithm and hash_value):
        print("Error: mutually dependant parameters: --hash-algorithm and --hash-value.", file=sys.stderr)
        return 1
    status = project_ops.add_download(project,
                                      env_spec_name=env_spec_name,
                                      env_var=filename_variable,
                                      url=download_url,
                                      filename=filename,
                                      hash_algorithm=hash_algorithm,
                                      hash_value=hash_value)
    if status:
        print(status.status_description)
        print("Added %s to the project file." % download_url)
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def remove_download(project_dir, env_spec_name, filename_variable):
    """Remove a download requirement from project and from file system."""
    project = load_project(project_dir)
    # we can remove a download even if prepare fails, so disable
    # printing errors in the frontend.
    with project.null_frontend():
        result = prepare_without_interaction(project, env_spec_name=env_spec_name, mode=PROVIDE_MODE_CHECK)
    status = project_ops.remove_download(project,
                                         env_spec_name=env_spec_name,
                                         env_var=filename_variable,
                                         prepare_result=result)
    if status:
        print(status.status_description)
        print("Removed {} from the project file.".format(filename_variable))
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def list_downloads(project_dir, env_spec_name):
    """List the downloads present in project."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1

    if project.downloads(env_spec_name):
        print("Downloads for project: {}\n".format(project_dir))
        console_utils.print_names_and_descriptions(project.download_requirements(env_spec_name), name_attr='title')
    else:
        print("No downloads found in project.")
    return 0


def main_add(args):
    """Start the download command and return exit status code."""
    return add_download(args.directory, args.env_spec, args.filename_variable, args.download_url, args.filename,
                        args.hash_algorithm, args.hash_value)


def main_remove(args):
    """Start the remove download command and return exit status code."""
    return remove_download(args.directory, args.env_spec, args.filename_variable)


def main_list(args):
    """Start the list download command and return exit status code."""
    return list_downloads(args.directory, args.env_spec)
